
/*===========================================================================
 * Scrinchy, a simple webserver.
 * Version 0.10
 * Copyright (C) 2005 by Zachary T. Smith.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *=========================================================================*/

/*---------------------------------------------------------------------------
 * Change log
 *
 * 2005 Jul 01: conception
 * 2005 Jul 02: added automatic parsing of http header
 * 2005 Jul 02: added logging of http header
 * 2005 Jul 04: added very basic PHP support (no cookies, no URL params)
 * 2005 Jul 10: added witty retort
 * 2005 Jul 19: improved log reporting.
 * 2005 Jul 20: added support for absolute path location for HTML etc.
 * 2005 Jul 24: added support for Perl
 * 2005 Jul 24: made PHP support optional
 * 2005 Jul 24: fixed zombie problem.
	=> 0.4
 * 2005 Jul 25: genericized preload mechanism, created preload_file etc.
	=> 0.5
 * 2005 Jul 25: added IP blocking for when hosts.deny is unreliable.
	=> 0.6
 * 2005 Jul 26: improved logging with file lock.
	=> 0.7
 * 2005 Jul 29: added log buffer, to reduce delays waiting for log file.
 * 2005 Aug 01: added automatic rejection of requests with "Authorization".
 * 2005 Aug 01: program now remembers specific attacker IPs and blocks them.
	=> 0.8
 * 2005 Aug 01: added blocking when too few params (typical of attackers).
 * 2005 Aug 01: parameter parser now produces parameter numbers.
 * 2005 Aug 01: added limit of only 1 active request at a time per IP,
 *		with one more queueable.
 * 2005 Aug 03: added load/save of attacker list to/from file.
	=> 0.9
 * 2005 Aug 04: added optional Python support.
 * 2005 Aug 04: made Perl support optional.
 * 2005 Aug 04: added proper file extension parsing.
 * 2005 Aug 05: expanded number of waiting requests to unlimited.
 * 2005 Aug 06: added checking for and validation of User-agent.
	=> 0.10
 * 2005 Aug 06: added ability to run Sy (Scrinchy stack language)
 * 2005 Aug 06: conception of simple Sy interpreter
 * 2005 Aug 06: added arithmetic and logic ops to Sy
 * 2005 Aug 06: added basic text attributes to Sy
 * 2005 Aug 06: added table creation functions to Sy
 * 2005 Aug 06: added style and DIV creation functions to Sy
	=> 0.11
 * 2005 Aug 08: added set-cookie command to Sy
 * 2005 Aug 08: added HTTP_COOKIE support
 * 2005 Aug 08: fixed tempfile not getting unlinked bug.
 *-------------------------------------------------------------------------*/

// #define SUPPORT_SY
#define SUPPORT_PERL
#define SUPPORT_PYTHON
// #define SUPPORT_PHP

#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>	// socket, bind ec
#include <sys/file.h>	// flock
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>	// tolower
#include <errno.h>
#include <signal.h>	// sigaction
#include <sys/wait.h>	// waitpid
#include <fcntl.h>	// file lock


#define VERSION "0.11"


typedef unsigned long IP;


enum {
	HTTP_ERR_MOVED = 301,
	HTTP_ERR_BAD_REQUEST = 400,
	HTTP_ERR_UNAUTH = 401,
	HTTP_ERR_FORBIDDEN = 403,
	HTTP_ERR_NOT_FOUND = 404,
	HTTP_ERR_INTERNAL = 500,
	HTTP_ERR_BAD_EXTN = 999
};
enum {
	EXITCODE_HTTPERR_MOVED = 60,
	EXITCODE_HTTPERR_BAD_REQUEST = 61,
	EXITCODE_HTTPERR_UNAUTH = 62,
	EXITCODE_HTTPERR_FORBIDDEN = 63,
	EXITCODE_HTTPERR_NOT_FOUND = 64,
	EXITCODE_HTTPERR_INTERNAL =65,
	EXITCODE_PORTSCAN=66,
	EXITCODE_PARAM_ATTACK=67,
	EXITCODE_FLOCK_FAILED=68,
	EXITCODE_NOMEM=69,
	EXITCODE_HTTPERR_OTHER = 70,
	EXITCODE_READ_ERROR= 71,
	EXITCODE_FEW_PARAMS=72,
	EXITCODE_HTTP_ATTACK=73,
	EXITCODE_BUFFER_ATTACK=74,
};

enum {
	HEADER_HAS_AUTHORIZATION=1,
	HEADER_BAD_USERAGENT=2,
};

enum { 
	EXTN_NONE=0,
	EXTN_HTML=1,
	EXTN_PL=2,
	EXTN_PY=3,
	EXTN_PHP=4,
	EXTN_JPG=5,
	EXTN_GIF=6,
	EXTN_PNG=7,
	EXTN_MPG=8,
	EXTN_AVI=9,
	EXTN_ICO=10,
	EXTN_MP3=11,
	EXTN_OGG=12,
	EXTN_JS=13,
	EXTN_SY=14,
};


enum {
	PARAM_HOST=1,
	PARAM_USERAGENT=2,
	PARAM_ACCEPT=3,
	PARAM_ACCEPTLANGUAGE=4,
	PARAM_ACCEPTENCODING=5,
	PARAM_ACCEPTCHARSET=6,
	PARAM_KEEPALIVE=7,
	PARAM_CONNECTION=8,
	PARAM_AUTHORIZATION=9,
	PARAM_REFERER=10,
	PARAM_CACHECONTROL=11,
	PARAM_COOKIE=12,
};

#define REQLEN 65536
#define MAXREQ 65500 /* reasonable maximum, we don't support cookies */

#ifndef MAXPATH
#define MAXPATH 4096
#endif


typedef char bool;
enum { false=0, true=1 };


typedef struct {
	char *name;
	char code;
	char *contents;
}
StringPair;


typedef struct li {
	pid_t	pid;
	IP	addr;
	struct li 	*next;
} 
ListItem;


typedef struct sli {
	int	fd;
	pid_t	pid;
	struct sli 	*next; 
} 
AddrListSubItem;

typedef struct li2 {
	IP	addr;
	AddrListSubItem	*list;
	struct li2 	*next;
} 
AddrListItem;


#ifdef SUPPORT_SY
typedef struct stackim {
	int	integer;
	double	real;
	char*	string;
	char	type;
	unsigned int bold, italic, underline, super, sub, strikeout;
	struct stackim	*next; 
} 
StackItem;

typedef struct namevaluepair {
	char 	*name;
	StackItem	value;
	struct namevaluepair *next;
} NameValuePair;

bool process_sy (int, char *);
bool process_sy_core (int, char *codepath, char *inpath, char *outpath,char*);
#endif



enum {
	STACKITEM_REAL=1,
	STACKITEM_INTEGER=2,
	STACKITEM_STRING=3,
	STACKITEM_SYMBOL=4,
	STACKITEM_CODE=5,
	STACKITEM_STYLE=5,
};


#define WEBDIR "/var/scrinchy/"
#define LOGFILE "/tmp/scrinchy.log"
#define ATTACKERS_FILE "/etc/scrinchy/attackers"


bool process_script (int fd, char *localpath, char type);


static int webdirlen;

static char header[2000];
static int headerlen = 0;

static unsigned char ip_block_all [256];
static short ip_block_sub1 [256];
static short ip_block_sub2 [256];

#define MAXATTACKERS 1024
static IP	*attackers;
static int total_attackers = 0;

#define IPBLOCK_ALL(num) { ip_block_all[num]=1; }
#define IPBLOCK_SUB1(high,num) { ip_block_sub1[high]=num; }
#define IPBLOCK_SUB2(high,num) { ip_block_sub1[high]=num; }

#define MAXHASH 256
#define HASHAND 255
static ListItem *pid_hash[MAXHASH];

static AddrListItem *addr_hash[256]; // key is last # of IP

#define MAXLOG 65536
static char my_log [MAXLOG];
static int my_log_ix = 0;


static char *sad_reality = "\
<h3>Did you know? \
The evidence shows that George W Bush stole the 2000 and 2004 elections \
thanks to rigging efforts by Neocon operatives \
in key states such as Florida, Ohio and New Mexico.<p>\
</h3>\
<div style=\"background: #f0f0ff;\">\
<b>In Florida</b>, his brother Jeb Bush managed the rigging operations in 2000 and \
2004.  Rigging Florida was easy: Jeb simply told 50,000 voters \
(virtually all of them Democrats, most black) that they \
were no longer allowed to vote, citing similarities \
between their names and those of out-of-state ex-felons. \
You see, in Florida, voters are not identified \
by Social Security number, but rather by name. Technically ex-felons \
<i>may</i> legally vote. \
The similarities were so specious, as Greg Palast has shown, \
that often age, race and gender difference were ignored. \
But of course, \
Jeb just needed a pretext to remove them from the rolls \
so that his brother would win. And George did, each time by slim margins. \
After the 2000 rig was uncovered, Jeb's lawbreaking was proven in court \
and the judge ordered Jeb Bush to \
reinstate the wronged voters. \
But in typical Neocon disdain for the law, he never did. \
<p>\
<b>In Ohio</b>, in 2004 Ken Blackwell, who was \
George W Bush's co-campaign manager for Ohio, \
was in fact in charge of overseeing the counting of votes for the state, \
because Blackwell was also Secretary of State. \
It was because of him that voting machines were plentiful \
in rich areas but ordered moved out of poor districts, \
where some people waited outside in the cold, even in the rain as long as 11 hours in line to vote. \
When it became clear that the 2004 election was fraudulent, \
activists raised the huge sum of money required for a vote recount. \
But the recount itself was then rigged as well, in some cases \
in view of cameras, because partisan Blackwell was in charge \
of it. \
<p>\
<b>In New Mexico</b>, it was said by activists that poor Hispanics \
were often given \"provisional\" ballots often in violation of rules, \
the result being that their votes were never counted. \
Activists seeking a recount in New Mexico were blocked from doing so, \
with the state government breaking its own laws to prevent the recount. \
Why was this? It could have something to do with the fact that \
Democratic governor Bill Richardson \
was formerly an oil executive at Valero in Texas. \
<p>\
<b>Electronic voting machines</b>: It is also worth pointing out that \
some two-thirds of Americans voted using electronic voting machines in 2004. \
In the USA, these machines produce no paper trail or receipt, \
and in fact they are manufactured by only 3 or 4 companies, the two largest \
of them (Diebold and ESS) being owned by ardent Bush supporting Republicans. \
One executive from Diebold, Walden O'Dell, \
proclaimed he was \
\"committed to helping Ohio deliver its electoral votes to the president next year.\".\
<p> \
<a href=http://www.electionfraud2004.org>link</a> \
<a href=http://www.blackboxvoting.org>link</a> \
<a href=http://www.votersunite.org/electionproblems.asp>link</a> \
<a href=http://news.bbc.co.uk/1/hi/programmes/newsnight/3956129.stm#>link</a> \
<a href=http://www.gregpalast.com/detail.cfm?artid=393&row=1>link</a> \
<a href=http://www.gregpalast.com>link</a>\n\
<a href=http://www.motherjones.com/commentary/columns/2004/03/03_200.html>link</a> \
<a href=http://www.commondreams.org/headlines03/0828-08.htm>link</a></div>\n";

static int sad_reality_len;

typedef struct param {
	char *name;
	char code;
} KnownParam;

static KnownParam known_params [] = {
	{ "Host", PARAM_HOST },
	{ "Referer", PARAM_REFERER },
	{ "User-Agent", PARAM_USERAGENT},
	{ "Accept", PARAM_ACCEPT },
	{ "Accept-Language", PARAM_ACCEPTLANGUAGE },
	{ "Accept-Encoding", PARAM_ACCEPTENCODING },
	{ "Accept-Charset", PARAM_ACCEPTCHARSET },
	{ "Keep-Alive", PARAM_KEEPALIVE },
	{ "Connection", PARAM_CONNECTION },
	{ "Authorization", PARAM_AUTHORIZATION },
	{ "Cache-control", PARAM_CACHECONTROL },
	{ "Cookie", PARAM_COOKIE },
	{ NULL, 0 }
};


int myexit(int code)
{
	// printf ("exit, code=%d\n", code);
	exit(code);
}

void log_store ()
{
	FILE *f;
	while (!(f = fopen ("/tmp/scrinchy.log", "a")))
		;
	int i = flock (f->_fileno, LOCK_EX);
	if (i) {
		perror ("flock");
		myexit(EXITCODE_FLOCK_FAILED); // flock should block
	}
	fwrite (my_log, 1, my_log_ix, f);
	fclose (f);

	my_log_ix = 0;
}


void log_write (char *s)
{
	if (!s) 
		return;

	int len = strlen(s);
	int remaining = MAXLOG - my_log_ix - 10;

	if (len > remaining) {
		strncpy (&my_log[my_log_ix], s, remaining);
		my_log_ix += remaining;
		my_log[my_log_ix] = 0;
	} else {
		strcpy (&my_log[my_log_ix], s);
		my_log_ix += len;
	}
}

void log_writeln_num (int n)
{
	char tmp[20];
	sprintf (tmp, "%2d\n", n);
	log_write (tmp);
}

void log_write_num (int n)
{
	char tmp[20];
	sprintf (tmp, "%2d", n);
	log_write (tmp);
}

void log_writeln (char *s)
{
	if (s)
		log_write (s);
	log_write ("\n");
}


void error (int which)
{
	fprintf (stderr, "Error! #%d\n", which);
	log_store ();
	myexit(32+which);
}

void warning (char *s)
{
	fprintf (stderr, "Warning! %s\n", s);
}


bool strcasestr (char *s, char *s2)
{
	if (!s || !s2)
		return false;

	while (*s) {
		if (tolower(*s) == tolower(*s2)) {
			if (!strcasecmp(s+1,s2+1))
				return true;
		}	
		s++;
	}
	return false;
}


char parse_extension (char *s)
{
	if (!s)
		return EXTN_NONE;

	char c= *s++;
	char c2,c3;
	switch (c) {
	case 'i':
		if (!strcasecmp ("co",s))
			return EXTN_ICO;
		break;
	case 'h':
		if (!strcasecmp ("tml",s) || !strcasecmp ("tm",s))
			return EXTN_HTML;
		break;
	case 'j':
		if (!strcasecmp ("pg",s) || !strcasecmp ("peg",s))
			return EXTN_JPG;
		c2 = tolower(*s++);
		if (c2=='s' && !*s)
			return EXTN_JS;
		break;
	case 'g':
		if (!strcasecmp ("if",s))
			return EXTN_GIF;
		break;
	case 'o':
		if (!strcasecmp ("gg",s))
			return EXTN_OGG;
		break;
	case 'a':
		if (!strcasecmp ("vi",s))
			return EXTN_AVI;
		break;
	case 'm':
		c2 = tolower(*s++);
		if (c2 == 'p') {
			char c3 = tolower(*s++);
			if (c3 == 'g' && !*s)
				return EXTN_MPG;
			else if (c3 == '3' && !*s)
				return EXTN_MP3;
		}
		break;
	case 'p':
		c2 = tolower(*s++);
		switch (c2) {
		case 'l':
			if (!*s)
				return EXTN_PL;
			break;
		case 'y':
			if (!*s)
				return EXTN_PY;
			break;
		case 'h': 
			c3=tolower(*s++);
			if (c3=='p' && !*s)
				return EXTN_PHP;
			break;
		case 'n':
			c3=tolower(*s++);
			if (c3=='g' && !*s)
				return EXTN_PNG;
			break;
		}
		break;
	case 's':
		c2 = tolower(*s++);
		if (c2=='y' && !*s)
			return EXTN_SY;
	}
	return EXTN_NONE;
}

char lookup_param (char *s)
{
	int i=0;
	char *s2;
	if (!s)
		return 0;
	while ((s2 = known_params[i].name)) {
		if (!strcasecmp (s2, s))
			return known_params[i].code;
		i++;
	}
	return -1;
}

void ip_to_str(IP addr, char *str)
{
	if (!str)
		return;

	unsigned short a = 0xff & addr;
	unsigned short b = 0xff & (addr >> 8);
	unsigned short c = 0xff & (addr >> 16);
	unsigned short d = 0xff & (addr >> 24);
	sprintf (str, "%d.%d.%d.%d", a,b,c,d);
}

	
short get_pid_hash_key (pid_t pid)
{
	if (pid < 0)
		return -1;

	return HASHAND & (unsigned)pid;
}

ListItem *pid_hash_remove (pid_t pid)
{
	if (pid < 0)
		return NULL;

	short key = get_pid_hash_key (pid);
	if (key < 0)
		return NULL;

	ListItem *im = pid_hash[key];
	ListItem *prev = NULL;
	bool found = 0;

	while (im) {
		if (im->pid == pid) {
			found = 1;
			break;
		}

		prev = im;
		im = im->next;
	}

	if (!found)
		return NULL;

	if (!prev) 
		pid_hash[key] = im->next;
	else 
		prev->next = im->next;

	return im;
}

AddrListItem *addr_hash_lookup (IP addr)
{
	short key = addr >> 24;

	AddrListItem *im = addr_hash[key];

	while (im) {
		if (im->addr == addr) {
			return im;
		}

		im = im->next;
	}
	return NULL;
}

AddrListItem *addr_hash_add (IP addr)
{
	unsigned short key = addr >> 24;

	AddrListItem *existing;
	existing = addr_hash_lookup (addr);
	if (existing)
		return existing;

	AddrListItem *nu = (AddrListItem*) malloc (sizeof (AddrListItem));
	if (!nu) {
		fprintf (stderr, "Out of memory\n");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	nu->list = NULL;
	nu->addr = addr;
	nu->next = addr_hash[key];
	addr_hash[key] = nu;
	return nu;
}


// If delayed processing, set pid = -1
// To update an already existing entry, just set the pid.
//
bool addr_hash_add_fd (IP addr, int fd, int pid)
{
	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		im = addr_hash_add (addr);
		if (!im) {
			return false;
		}
	}

	AddrListSubItem *sim;
	char tmp[30];
	ip_to_str (addr,tmp);

	sim = malloc (sizeof(AddrListSubItem));
	if (!sim)
		return false;

	sim->fd = fd;
	sim->pid = pid;
	sim->next = im->list;
	im->list = sim;
	return true;
}


// Find out how many requests we are currently processing 
// for a given IP.
int addr_hash_count_active (IP addr)
{
	int count_active=0;
	int count_inactive=0;

	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		return -1;
	}

	AddrListSubItem *sim = im->list;
	if (!sim) {
		warning ("AddrListItem exists without list");
	}

	while (sim) {
		if (sim->pid != -1) {
			count_active++;
		}
		else
			count_inactive++;
		sim = sim->next;
	}

#if 0
char tmp[30];
ip_to_str (addr,tmp);
printf ("found that IP %s has %d active connections, %d inactive\n",
	tmp,count_active,count_inactive);
#endif

	return count_active;
}


	
// Remove all of the entries for a given IP.
//
int addr_hash_remove_all (IP addr)
{
	unsigned short key = addr >> 24;

	AddrListItem *im = addr_hash[key];
	AddrListItem *prev = NULL;

	while (im) {
		if (im->addr == addr) {
			break;
		}

		prev = im;
		im = im->next;
	}

	if (!im) {
		return -1;
	}

	if (!prev)
		addr_hash[key] = im->next;
	else 
		prev->next = im->next;

	// Remove all the connections.
	AddrListSubItem *sim = im->list;
	int count=0;
	while (sim) {
		count++;
		AddrListSubItem *next = sim->next;
		close (sim->fd);
		free (sim);
		sim = next;
	}
	im->list = NULL;
	free (im);

	return count;
}

	
AddrListSubItem *addr_hash_remove_one (IP addr, int pid)
{
	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		return NULL;
	}

	AddrListSubItem *prev = NULL;
	AddrListSubItem *sim = im->list;
	if (!sim) {
		return NULL;
	}
	while (sim && pid != sim->pid) {
		prev = sim;
		sim = sim->next;
	}
	if (!sim) {
		return NULL;
	}
	if (prev) {
		prev->next = sim->next;
	} else {
		im->list = sim->next;
	}
	sim->next = NULL;

#if 0
char tmp[30];
ip_to_str (addr,tmp);
printf ("removed IP %s, pid %d\n", tmp,sim->pid);
#endif

	if (!im->list)
		addr_hash_remove_all (addr);

	return sim;
}

void pid_hash_add (pid_t pid, IP addr)
{
	if (pid < 0)
		return;

	short key = get_pid_hash_key (pid);
	if (key < 0)
		return;

	ListItem *tmp;
	if ((tmp = pid_hash_remove (pid))) {
		fprintf (stderr, "Process id already in pid hash!\n");
	}

	ListItem *nu = (ListItem*) malloc (sizeof (ListItem));
	if (!nu) {
		fprintf (stderr, "Out of memory\n");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	nu->pid = pid;
	nu->addr = addr;
	nu->next = pid_hash[key];
	pid_hash[key] = nu;
}



void maketempfile (char *path)
{
	strcpy (path, "/tmp/_scrinchy_XXXXXX");
	mkstemp (path);
}

int process_request (int fd, struct sockaddr_in *from);


/* In-memory cacheing of specific files.
 */
#define MAXPRELOAD 8
static unsigned char *preload [MAXPRELOAD];
static int preload_len [MAXPRELOAD];
enum {
	PRELOAD_INDEX = 0,
	PRELOAD_FAVICON = 1
};




bool load_attackers ()
{
	FILE *f = fopen (ATTACKERS_FILE, "r");
	if (!f) {
		log_writeln ("Unable to read attackers list.");
		log_store ();
		return false;
	}

	int count=0;
	if (1 != fscanf(f, "%d", &count)) {
		log_writeln ("Bad format in attackers list.");
		log_store ();
		fclose(f);
		return false;
	}

	if (count > MAXATTACKERS)
		count = MAXATTACKERS;
	total_attackers=count;

	int i;
	for (i=0; i<count; i++)
		attackers[i]=0;

	for (i=0; i<count; i++) {
		unsigned long a,b,c,d;
		if (4 != fscanf (f, "%lu.%lu.%lu.%lu", &a,&b,&c,&d)) {
			log_writeln ("Bad format in attackers list.");
			log_store ();
			fclose(f);
			total_attackers = count;
			return false;
		}
		a &= 0xff;
		b &= 0xff;
		c &= 0xff;
		d &= 0xff;
		b <<= 8;
		c <<= 16;
		d <<= 24;
		attackers[i] = a|b|c|d;
	}

	log_writeln (NULL);
	log_write ("Total attacker IPs read from file: ");
	log_writeln_num (total_attackers);
	log_writeln (NULL);
	log_store ();

	return true;
}


bool save_attackers ()
{
	FILE *f = fopen (ATTACKERS_FILE, "w");
	if (!f) {
		log_writeln ("Unable to write attackers list.");
		log_store ();
		return false;
	}

	char ipstr[30];
	int i;
	fprintf (f, "%d\n", total_attackers);
	for (i=0; i<total_attackers; i++) {
		ip_to_str (attackers[i], ipstr);
		fprintf (f, "%s\n", ipstr);
	}
	fclose (f);
	return true;
}

void add_attacker (IP addr)
{
	if (total_attackers >= MAXATTACKERS) {
		log_write ( "too many attackers\n");
		log_store ();
		return;
	}

	attackers [total_attackers++] = addr;
	save_attackers();
}


unsigned char *get_file (char *path, int *lenreturn, bool prepend)
{
	struct stat st;
	char fullpath [MAXPATH];

	if (!path)
		return NULL;
	/*----------*/

	if (prepend) {
		strcpy (fullpath, WEBDIR);	// must include trailing '/'
		strncpy (fullpath+webdirlen, path, MAXPATH-2-webdirlen);
	} else 
		strncpy (fullpath, path, MAXPATH-2);

	fullpath[MAXPATH-1] = 0;

	// Get file length.
	if (stat (fullpath, &st)) {
		perror ("stat");
		return NULL;
	}
	int file_len = st.st_size;

	// Allocate buffer.
	unsigned char *data = (char*) malloc (file_len+1);
	if (!data) {
		perror ("malloc");
		log_writeln ("out of memory");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	// Read the entire file in one go.
	FILE *f = fopen (fullpath, "rb");
	if (!f)
		return NULL;
	int i = fread (data, 1, file_len, f);
	fclose (f);
	if (i != file_len)  {
		free (data);
		return NULL;
	}

	// Done. Return length.
	if (lenreturn)
		*lenreturn = file_len;

	// Store an end-of-string just in case.
	data[file_len] = 0;

	// Return file data.
	return data;
}

int preload_file (char *path, int which)
{
	int len;
	char *s;
	if (which < 0 || which >= MAXPRELOAD)
		return false;
	if (preload[which])
		return false;
	s = preload[which] = get_file (path, &len, 1);
	if (!s) {
		return false;
	}
	preload_len [which] = len;
	return true;
}


int parse_http_header (char *s, StringPair results[], int max_results,
			int *referer_index_return, int *flags_return)
{
	int count=0;

	if (!s) 
		return 0;
	if (max_results < 1) 
		return 0;
	if (!flags_return)
		return 0;
	/*---------*/

	if (flags_return)
		*flags_return = 0;

	// Skip the get/post
	char *foo = strchr (s, '\n');
	if (!foo)
		foo = strchr (s, '\r');
	if (!foo)
		return 0;
	s = foo+1;
	if (*s == '\n' || *s == '\r')
		s++;

	// Keep track of User-agent
	char found_useragent=false;

	// Process the data
	while (*s && count < max_results) 
	{
		while (*s && (*s==' ' || *s=='\t' || *s=='\r'))
			s++;

		char *s2 = strchr (s, '\n');
		if (!s2)
			s2 = strchr (s, '\r');
		if (!s2)
			break;
		*s2 = 0;

		char *colon = strchr (s, ':');
		if (!colon)
			break;
		*colon = 0;

		char *contents = colon+1;
		while (*contents && (*contents==' ' || *contents=='\t'))
			contents++;

		char *junk = strchr (contents, '\r');
		if (junk)
			*junk=0;

		char code = lookup_param (s);
		results[count].name = strdup (s);
		results[count].code = code;
		if (code < 0) {
			log_write ("Unknown param: ");
			log_writeln (s);
		}
		if (code == PARAM_REFERER)
			*referer_index_return = count;
		else
		if (flags_return && code == PARAM_AUTHORIZATION)
			*flags_return |= HEADER_HAS_AUTHORIZATION;
		else
		if (code == PARAM_USERAGENT) {
			if (strcasestr (s, "Dillo") ||
			    strcasestr (s, "Lynx") ||
			    strcasestr (s, "Mozilla") ||
			    strcasestr (s, "Explorer") ||
			    strcasestr (s, "Netscape"))
				found_useragent = true;
		}
		else
		if (code == PARAM_COOKIE) {
			setenv ("HTTP_COOKIE", contents, 1);
		}

		results[count].contents = strdup (contents);
		count++;

		s = s2+1;
	}

	if (!found_useragent)
		*flags_return |= HEADER_BAD_USERAGENT;

	return count;
}



int stat_file (char *path)
{
	if (!path)
		return -1;

	struct stat st;
#if 0
	char fullpath[MAXPATH];
	strcpy (fullpath, WEBDIR);
	strncpy (fullpath+webdirlen, path, MAXPATH-2-webdirlen);
	fullpath[MAXPATH-1] = 0;
	setenv ("PATH_INFO", fullpath, 1);
	setenv ("PATH_TRANSLATED", fullpath, 1);
#endif
	if (stat (path, &st)) {
		perror ("stat");
		return -1;
	}

	return st.st_size;
}


int main (int argc, char **argv)
{
	int i;
	int listensock;
	int clientsock;
	struct servent *sp;
	struct sockaddr_in sin, from;

	my_log_ix = 0;

	log_writeln(NULL);
	log_writeln ("=============================================================================");
	log_write ("Scrinchy server started @ ");
	time_t t = time(NULL);
	log_write (ctime(&t)); // ctime appends newline
	log_writeln ("=============================================================================");
	log_store ();

	attackers = (IP*) malloc(MAXATTACKERS * 4);
	if (!attackers) {
		fprintf (stderr, "out of memory");
		return 0;
	}
	load_attackers();

	sad_reality_len = strlen (sad_reality);

	// Clear the IP blocking array.
	for (i=0; i<256; i++) {
		ip_block_all [i] = 0;
		ip_block_sub1[i] = -1;
		ip_block_sub2[i] = -1;
	}

	for (i=0; i<MAXHASH; i++)
		pid_hash[i] = NULL;

	for (i=0; i<256; i++)
		addr_hash[i] = NULL;

	// Lock out the hackers
	// This is necessary because hosts.deny has proven ineffective.
	IPBLOCK_ALL (201);
	IPBLOCK_ALL (202);
	IPBLOCK_ALL (203);
	IPBLOCK_ALL (204);
	IPBLOCK_ALL (205);
	IPBLOCK_ALL (206);
	IPBLOCK_ALL (207);
	IPBLOCK_ALL (208);
	IPBLOCK_ALL (209);
	IPBLOCK_ALL (210);
#if 1
	IPBLOCK_ALL (211);
	IPBLOCK_ALL (212);
	IPBLOCK_ALL (213);
	IPBLOCK_ALL (214);
	IPBLOCK_ALL (215);
	IPBLOCK_ALL (216);
	IPBLOCK_ALL (217);
	IPBLOCK_ALL (218);
	IPBLOCK_ALL (219);
	IPBLOCK_ALL (220);
	IPBLOCK_ALL (221);
	IPBLOCK_ALL (222);
	IPBLOCK_SUB1(80,17);
	IPBLOCK_SUB1(60,36);
	IPBLOCK_SUB1(195,70);
#endif

	// Set up boilerplate header.
	//
	sprintf (header, "\
HTTP/1.0 200 OK\r\n\
Server: Scrinchy version %s\n\
Connection: close\n", VERSION);
	headerlen = strlen (header);

	// Time saver.
	webdirlen = strlen(WEBDIR);

#if 0
	// Ignore child handler.
	//
	signal(SIGCHLD, SIG_IGN);
#endif

	// Read in the index to improve speed.
	//
	preload_file ("index.html", PRELOAD_INDEX);
	preload_file ("favicon.ico", PRELOAD_FAVICON);

	// Banner
	//
	printf ("\nThis is Scrinchy webserver, version %s\n", VERSION);
	printf ("Copyright (C) 2005 by Zack T Smith.\n\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
file COPYING, which contains the GNU General Public License,\n\
for more details.\n\n");

	char *service = "http";

	if ((sp = getservbyname (service, "tcp")) == NULL) {
		perror ("getservbyname");
		error (1);
	}

	sin.sin_family = AF_INET;
	sin.sin_port = sp->s_port;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);

	if (( listensock = socket (AF_INET, SOCK_STREAM, 0)) < 0)  {
		perror ("socket");
		error (2);
	}

	if (bind (listensock, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
		perror ("bind");
		error (3);
	}

	if (listen (listensock, 500) < 0) {
		perror ("listen");
		error (4);
	}

	// Main loop
	//
	while(1) {
		// Wait for a client to contact us.
		socklen_t len;
		clientsock = accept(listensock, 
				(struct sockaddr*) &from, &len);
		if (clientsock < 0) {
			perror ("accept");
			error (5);
		}

		if (len != sizeof (struct sockaddr_in))
			error (6);
		
		IP addr = from.sin_addr.s_addr;

		char found_attacker=0;
		i=0;
		while (!found_attacker && i < total_attackers) {
			IP m = attackers[i++];
			if (addr != m)
				continue;

			found_attacker=1;
		}
		if (found_attacker) {
			close(clientsock);
			continue;
		}

		// Here we shall refuse to deal with more than one
		// request at a time from any given IP, because one form 
		// of attack involves making many simultaneous requests.
		//
		bool dont_execute = false;
		int count = addr_hash_count_active (addr);
		if (count >= 1) {
			dont_execute = 1;
			addr_hash_add_fd (addr, clientsock, -1);
		}

		AddrListSubItem *pending=NULL;
process_more:	if (pending)
			dont_execute=false;

		// Create a child process to deal with request.
		if (!dont_execute) {
			int pid;

			if (!(pid = fork())) {
				my_log_ix = 0;
				process_request (clientsock, &from);
				myexit (0);
			}

			// Let's keep track of the child
			pid_hash_add (pid, addr);

			if (!pending)
				addr_hash_add_fd (addr, clientsock, pid);
			else
				pending->pid = pid;

			// Release parent copy of socket.
			close (clientsock);
		}

		// Check up on children that have exited
		int child, status;
		while ((child = waitpid (-1, &status, WNOHANG))) {
			if (child < 0) {
perror ("waitpid");
				break;
			}

			if (WIFEXITED (status)) {
				status = WEXITSTATUS (status);
			} else
				status = -1;

			ListItem *im = pid_hash_remove (child);
			if (!im) {
				fprintf (stderr, "Failed to remove child pid %d from hash\n", child);
				continue;
			}

//HERE IS WHERE THE SEGFAULT OCCURRED
			AddrListSubItem *sim = addr_hash_remove_one (im->addr, 
					child);
			if (!sim) {
				fprintf (stderr, "Failed to remove addr %08lx, pid %d from hash\n", im->addr,child);
				continue;
			}
			free(sim);

			char ipstring[30];
			ip_to_str (im->addr, ipstring);

#if 0
			// Debugging stuff!
			FILE *ff = stdout;
			fprintf (ff, "ip=%s\t", ipstring);
			switch (status) {
			case EXITCODE_HTTPERR_MOVED:
				fprintf (ff, "http: moved\n");
				break;
			case EXITCODE_READ_ERROR:
				fprintf (ff, "read error\n");
				break;
			case EXITCODE_HTTPERR_BAD_REQUEST:
				fprintf (ff, "http: bad request\n");
				break;
			case EXITCODE_HTTPERR_UNAUTH:
				fprintf (ff, "http: unauthorized\n");
				break;
			case EXITCODE_HTTPERR_FORBIDDEN:
				fprintf (ff, "http: forbidden\n");
				break;
			case EXITCODE_HTTPERR_NOT_FOUND:
				fprintf (ff, "http: not found\n");
				break;
			case EXITCODE_HTTPERR_INTERNAL:
				fprintf (ff, "http: internal\n");
				break;
			case EXITCODE_HTTPERR_OTHER:
				fprintf (ff, "http: other\n");
				break;
			case EXITCODE_PORTSCAN:
				fprintf (ff, "port scanner\n");
				break;
			case EXITCODE_PARAM_ATTACK:
				fprintf (ff, "parameter attack\n");
				break;
			case EXITCODE_BUFFER_ATTACK:
				fprintf (ff, "buffer attack\n");
				break;
			case EXITCODE_FEW_PARAMS:
				fprintf (ff, "too few parameters\n");
				break;
			case EXITCODE_FLOCK_FAILED:
				fprintf (ff, "flock failed\n");
				break;
			case EXITCODE_NOMEM:
				fprintf (ff, "out of memory\n");
				break;
			case EXITCODE_HTTP_ATTACK:
				fprintf (ff, "http:// atack\n");
				break;
			default:
				fprintf (ff, "other result %d\n", status);
			}
			//fclose(ff);
#endif

			// Do we have a attacker?
			if (status == EXITCODE_PORTSCAN ||
			    status == EXITCODE_PARAM_ATTACK ||
			    status == EXITCODE_FEW_PARAMS ||
			    status == EXITCODE_HTTP_ATTACK ||
			    status == EXITCODE_BUFFER_ATTACK ||
			    status == EXITCODE_READ_ERROR) 
			{
				log_write ("Probable attacker at ");
				log_writeln (ipstring);
				add_attacker (im->addr);
				int num = addr_hash_remove_all (im->addr);
				log_write ("Removed ");
				log_write_num (num);
				log_writeln (" pending connections.");
				log_store ();
			} else {
				AddrListItem *im2 = addr_hash_lookup (im->addr);
				if (im2 && im2->list) {
					// Just grab the first one in the list.
					addr = im->addr;
					clientsock = im2->list->fd;
					pending = im2->list;
					//printf ("pending fd=%d\n",clientsock);
					free (im);
					goto process_more;
				} else {
					//printf ("no pending connection\n");
				}
			}

			free (im);
		}
	}

	close (listensock);
	free (attackers);
	return 0;
} // main()


int http_error (int fd, int which)
{
	char str[2000];

	log_store ();

	sprintf(str, "HTTP/1.0 %d\r\nConnection: close\r\n\r\n", which);
	write (fd, str, strlen(str));
	write (fd, "\r\n", 2);
	char *s = NULL;
	switch (which) {
	case HTTP_ERR_NOT_FOUND: 
		s = "File not found"; 
		break;
	case HTTP_ERR_MOVED: 
		s = "File has moved"; 
		break;
	case HTTP_ERR_BAD_REQUEST: 
		s = "Bad request"; 
		break;
	case HTTP_ERR_UNAUTH: 
		s = "Unauthorized"; 
		break;
	case HTTP_ERR_FORBIDDEN: 
		s = "Forbidden / Vietato / Verboten / Interdit"; 
		break;
	case HTTP_ERR_INTERNAL: 
		s = "Internal error"; 
		break;
	case HTTP_ERR_BAD_EXTN:
		s = "File extension not supported"; 
		break;
	default: s = "??";
	}

	sprintf (str, "<html>\
<head>\
<body alink=yellow vlink=green link=green bgcolor=#f8f8f8>\n \
<font color=#000080><p>\n\
<table width=100%% border=0>\
<tr>\
<td bgcolor=#e0e0ff>\n\
<font color=black size=+3><b> An error has occurred:\n\
</b><i><small> %s (number %d)\n</small></i></td></tr></table>\n\n", s, which);

	write (fd, str, strlen(str));
	write (fd, sad_reality, sad_reality_len);

	close (fd);

	switch (which) {
	case HTTP_ERR_NOT_FOUND: 
		myexit(EXITCODE_HTTPERR_NOT_FOUND);
	case HTTP_ERR_MOVED: 
		myexit(EXITCODE_HTTPERR_MOVED);
	case HTTP_ERR_BAD_REQUEST: 
		myexit(EXITCODE_HTTPERR_BAD_REQUEST);
	case HTTP_ERR_UNAUTH: 
		myexit(EXITCODE_HTTPERR_UNAUTH);
	case HTTP_ERR_FORBIDDEN: 
		myexit(EXITCODE_HTTPERR_FORBIDDEN);
	case HTTP_ERR_INTERNAL: 
		myexit(EXITCODE_HTTPERR_INTERNAL);
	default: 
		myexit(EXITCODE_HTTPERR_OTHER);
	}
	return 0;
}


int send_feces (int fd)
{
	char *feces = "<html><body><font size=+3>Your IP looks like that of a hacker, thus I repel you.";
	int feces_len = strlen (feces);

	char lengthstr[200];
	sprintf (lengthstr, "Content-length: %d\r\n", feces_len);

	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);
	
	int len = write (fd, feces, feces_len);
	close (fd);

	bool success = len == feces_len;

	log_write ("   \\\n    -> Sent feces");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(EXITCODE_HTTP_ATTACK);
}

int send_witty (int fd)
{
	char *witty = "<html><body><font size=+3>Only a girly-man would make a request such as you have just made.</html>";
	int wittylen = strlen(witty);

	char lengthstr[200];
	sprintf (lengthstr, "Content-length: %d\r\n", wittylen);
	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);
	
	int len;
	len = write (fd, witty, wittylen);
	close (fd);

	bool success = len == wittylen;

	log_write ("   \\\n    -> Sent wittiness");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(0);
}

int send_preload (int fd, int which)
{
	char lengthstr[200];

	if (which < 0 || which >= MAXPRELOAD)
		myexit(1);
	if (!preload [which] || preload_len[which]<=0)
		myexit(2);

	sprintf (lengthstr, "Content-length: %d\r\n", preload_len[which]);
	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);

	int len;
	len = write (fd, preload[which], preload_len[which]);
	close (fd);

	switch (which) {
	case PRELOAD_INDEX: 
		log_write ("   \\\n    -> Sent INDEX"); 
		break;
	case PRELOAD_FAVICON: 
		log_write ("   \\\n    -> Sent FAVICON"); 
		break;
	default:
		log_write ("   \\\n    -> Sent preload"); 
		break;
	}

	bool success = len == preload_len[which];

	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(1);
}

int send_file (int fd, char *path)
{
	if (!path) 
		return false;

	int len;
	unsigned char *data = get_file (path, &len, 0);
	if (!data)
		return false;

	int len2;
	len2 = write (fd, data, len);
	close (fd);

	bool success = len == len2;

	log_write ("   \\\n    -> Sent file '"); 
	log_write (path);
	log_write ("'");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(1);
}


// Send the results of running a script
// Optionally precede by some header stuff
//
int send_results (int fd, char *path, char *headerpath)
{
	if (!path) {
		if (headerpath)
			unlink (headerpath);
		return false;
	}
	/*----------*/

	int len = 0;
	unsigned char *data = get_file (path, &len, 0);
	unlink (path);
	if (!data)
		return false;

	int hdrlen;
	unsigned char *hdr=NULL;
	if (headerpath) {
		hdr = get_file (headerpath, &hdrlen, 0);
		if (headerpath)
			unlink (headerpath);
		if (!hdr)
			return false;
	}
	bool success = true;

	if (headerlen != write (fd, header, headerlen))
		success = false;
	else
	{
		if (headerpath) {
			if (hdrlen != write (fd, hdr, hdrlen))
				success = false;
			else
			if (2 != write (fd, "\n\n", 2))
				success = false;
		}

		if (success) {
			if (len != write (fd, data, len))
				success = false;
		}
	}
	close (fd);

	log_write ("   \\\n    -> Sent results (");
	log_write_num (len);
	log_write (") ");
	if (headerpath)
		log_write ("with extra header");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return true;
}


int process_request (int fd, struct sockaddr_in *from)
{
	static unsigned char req [REQLEN];
	static unsigned char ipstr [20];
	int i;

	if (!from || fd<0)
		myexit(98);
	/*----------*/

	IP addr = from->sin_addr.s_addr;
	unsigned short prefix = 0xff & addr;
	unsigned short second = 0xff & (addr >> 8);
	unsigned short third = 0xff & (addr >> 16);
	unsigned short fourth = 0xff & (addr >> 24);
	time_t t = time(NULL);

	sprintf (ipstr, "%d.%d.%d.%d", prefix, second, third, fourth);

#if 0
	struct hostent *he = gethostbyaddr (from, sizeof(from), AF_INET);
	char *remote_host = he ? he->h_name : "???";
#else
	char *remote_host = "unknown";
#endif

	log_writeln ("\n-------------------------------------------------------------------------------");
	log_write ("Request from ");
	log_write (ipstr);
	log_write (", on ");
	log_writeln (ctime(&t));

	i = read (fd, req, REQLEN);
	if (i >= 0)
		req[i]=0;

	log_write ("Bytes read: ");
	log_writeln_num (i);

	/* Are they on our shitlist? If so, send feces.
	 */
	if (ip_block_all[prefix] ||
	    ip_block_sub1[prefix] == second ||
	    ip_block_sub2[prefix] == second)
	{
		char *s = strchr (req,'\n');
		if (s) {
			*s = 0;
			log_write ("Request from blocked IP was: ");
			log_writeln (req);
		}
		send_feces (fd);
	}

	if (i >= MAXREQ) {
		log_writeln ("\tERROR: request too long");
		log_store ();
		myexit(EXITCODE_BUFFER_ATTACK);
	}
	else
	if (i < 0) {
		log_write ("\tERROR: read error\n");
		log_store ();
		perror ("read");
		myexit(EXITCODE_READ_ERROR);
	}
	else
	if (i < 4) {
		/* In my experience, it is possible that a browser
		 * can generate a zero-byte request when the user
		 * repeatedly reloads. It is therefore best to call
		 * this a "bad request" rather than an attack.
		 */
		log_write ("\tERROR: very short request\n");
		log_store ();
		http_error (fd, HTTP_ERR_BAD_REQUEST);
	}

	req[i]=0;
	if (strncasecmp (req, "get ", 4)) // no post yet
	{
		req[6] = 0;
		log_write ("\tERROR: non-GET request ");
		log_writeln (req);
		http_error (fd, HTTP_ERR_BAD_REQUEST);
	}

	// Parse the parameters
	StringPair pairs[50];
	int referer = -1;
	int param_flags = 0;
	int total_pairs = parse_http_header (req, pairs, 50, &referer,
		&param_flags);

	// Log the parameters
	if (!total_pairs)
		log_write ("\tno params\n");
	else {
		for (i=0; i<total_pairs; i++) {
			log_write ("Param ");
			log_write_num (i);
			log_write (": ");
			log_write (pairs[i].name);
			log_write (" = ");
			log_writeln (pairs[i].contents);
		}
	}

	if (total_pairs >= 40) {
		log_write ("\tERROR: improbably large number of params\n" );
		log_store ();
		myexit(EXITCODE_PARAM_ATTACK);
	}
	else if (total_pairs < 3) {
		log_write ("\tERROR: improbably few number of params\n" );
		log_store ();
		myexit(EXITCODE_FEW_PARAMS);
	}

	// Check for bad or missing parameters
	if (param_flags & HEADER_HAS_AUTHORIZATION) {
		log_write ("\tERROR: unnecessary Authorization included\n" );
		log_store ();
		myexit(EXITCODE_PARAM_ATTACK);
	}
	if (!param_flags & HEADER_BAD_USERAGENT) {
		log_write ("\tERROR: bad or missing User-agent\n" );
		log_store ();
		myexit(EXITCODE_PARAM_ATTACK);
	}

	// Extract the GET target.
#define MAXTARGETLEN 4096
	char target [MAXTARGETLEN + 2];
	int ix=0;
	char *s = req + 4;
	char *s2 = target;
	int ch;
	while (ch=*s, ch && (ch==' ' || ch=='\t'))
		s++;
	while (ix < MAXTARGETLEN) {
		ch = *s++;
		ix++;
		if (ch=='\n' || !ch || ch==' ' || ch=='\t')
			break;
		*s2++ = ch;
	}
	*s2 = 0;

	// Print target
	log_write ("GET target = '");
	log_write (target);
	log_writeln ("'");

	// Isolate any parameters
	char *params = strchr (target, '?');
	if (params) {
		*params = 0;
		params++;
	}

	// Provide the index
	//
	if ('/' == target[0] && !target[1] && index) 
		send_preload (fd, PRELOAD_INDEX);
	if ('/' == target[0] && tolower(target[1])=='i' && index) {
		if (!strncasecmp (target+2, "ndex.htm", 8))
			send_preload (fd, PRELOAD_INDEX);
	}
	if (strstr (target,"/favicon.ico"))
		send_preload (fd, PRELOAD_FAVICON);

	// Check for bizarre "http://" request
	if (!strncasecmp ("http:/", target, 6)) {
		send_witty (fd);
	}
	if (!strncasecmp ("/http:/", target, 7)) {
		send_witty (fd);
	}

	// Determine the extension if any.
	char *extension=NULL;
	ix = strlen (target) - 1;
	ch=0;
	while (ch = target[ix], ix > 0 && ch != '/' && ch != '.') {
		ix--;
	}
	if (ch == '.') {
		extension = &target[ix+1];
	}

	char extncode = 0;
	if (extension && *extension)
		extncode = parse_extension(extension);

	char localpath[MAXPATH];
	memset (localpath, 0, MAXPATH);

	// Clear the cookies
	setenv ("HTTP_COOKIE", "", 1);

	setenv ("REQUEST_METHOD", "GET", 1);
	setenv ("QUERY_STRING", params?params:"noparams=nothing", 1);

	switch (extncode) {
#ifdef SUPPORT_PERL
	case EXTN_PL:
		setenv ("DOCUMENT_ROOT", "/var/scrinchy/perl", 1);
		strcat (localpath, "/var/scrinchy/perl/");
		break;
#endif
#ifdef SUPPORT_PYTHON
	case EXTN_PY:
		setenv ("DOCUMENT_ROOT", "/var/scrinchy/python", 1);
		strcat (localpath, "/var/scrinchy/python/");
		break;
#endif
#ifdef SUPPORT_PHP
	case EXTN_PHP:
		setenv ("DOCUMENT_ROOT", "/var/scrinchy/php", 1);
		strcat (localpath, "/var/scrinchy/php/");
		break;
#endif
#ifdef SUPPORT_SY
	case EXTN_SY:
		setenv ("DOCUMENT_ROOT", "/var/scrinchy/sy", 1);
		strcat (localpath, "/var/scrinchy/sy/");
		break;
#endif

	case EXTN_NONE:
		log_write ("\tERROR: no extension\n");
		http_error (fd, HTTP_ERR_NOT_FOUND);

	default:
		strcat (localpath, "/var/scrinchy/");
	}

	setenv ("GATEWAY_INTERFACE", "CGI/1.1", 1);
	setenv ("HTTP_ACCEPT", "text/html", 1);
	setenv ("SERVER_PROTOCOL", "HTTP/1.0", 1);
	setenv ("SERVER_SOFTWARE", "Scrinchy", 1);
	setenv ("REMOTE_ADDR", ipstr, 1);
	setenv ("REMOTE_HOST", remote_host, 1);
	setenv ("HTTP_COOKIES", "", 1);
	setenv ("HTTP_REFERER", referer!=-1 ? pairs[referer].contents :
		"unknown", 1);

	// For now, some bogus stuff..
	setenv ("SERVER_NAME", "a.b.com", 1);
	setenv ("REMOTE_PORT", "9999", 1);
	setenv ("REMOTE_IDENT", "???", 1);

	if (*target == '/')
		strncat (localpath, target+1, 4000);
	else
		strncat (localpath, target, 4000);

	//----------------------------------------------
	// Verify that the HTML or script file exists.
	//
	int len = stat_file (localpath);
	if (len < 0) {
		log_write ("\tERROR: file not found\n");
		http_error (fd, HTTP_ERR_NOT_FOUND);
	}

	// Once target file is known to exist, process it.
	//
	switch (extncode) 
	{
#ifdef SUPPORT_PERL
	case EXTN_PL:
		if (!process_script (fd, localpath, 1))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_PYTHON
	case EXTN_PY:
		if (!process_script (fd, localpath, 3))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_PHP
	case EXTN_PHP:
		if (!process_script (fd, localpath, 2))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_SY
	case EXTN_SY:
		if (!process_sy (fd, localpath))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

	case EXTN_HTML:
	case EXTN_JPG:
	case EXTN_GIF:
	case EXTN_PNG:
	case EXTN_MPG:
	case EXTN_AVI:
	case EXTN_MP3:
	case EXTN_OGG:
	case EXTN_JS:
		if (!send_file (fd, localpath)) {
			log_write ("\tERROR: file send failed\n");
			http_error (fd, HTTP_ERR_NOT_FOUND);
		} else {
			log_store ();
			myexit(0);
		}
		return 0;

	default:
		http_error (fd, HTTP_ERR_BAD_EXTN);
	}
	return 0;
}


#ifdef SUPPORT_SY
int getword (FILE *f, char *buf, int max)
{
	if (!f || !buf)
		return -1;
	/*---------*/

	int ix=0;
	int ch = fgetc(f);
	while (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
		ch = fgetc(f);

	bool in_quote = ch == '"';
	if (in_quote)
		ch = fgetc(f);

	if (ch == EOF)
		return -1;
	buf[ix++] = ch;
	ch = fgetc(f);
	while (in_quote||(ch != ' ' && ch != '\t' && ch != '\r' && ch != '\n'))
	{
		if (ch == EOF)
			break;
		if (in_quote && ch == '"') {
			break;
		}
		if (ix < (max-1))
			buf[ix++] = ch;
		ch = fgetc(f);
	}
	buf[ix] = 0;

	if (!ix && ch==EOF)
		return -1;
	return ix;
}


StackItem *StackItem_new () 
{
	StackItem *nu = (StackItem*) malloc(sizeof(StackItem));
	if (!nu)
		return NULL;
	memset(nu,0,sizeof(StackItem));
	return nu;
}

void StackItem_delete (StackItem *im)
{
	if (im) {
		if (im->type == STACKITEM_STRING || 
		    im->type == STACKITEM_STYLE ||
		    im->type == STACKITEM_SYMBOL)
			free (im->string);
		free (im);
	}
}

bool process_sy_core (int fd, char *codepath, char *inpath, char *outpath,
	char *headerpath)
{
	FILE *inf, *ouf, *f, *hef;
	StackItem *stack = NULL;
	NameValuePair *vars = NULL;

	if (!inpath || !outpath)
		return false;
	/*---------*/

	f = fopen(codepath, "rb");
	if (!f) {
		perror ("fopen read");
		return false;
	}
	inf = fopen(inpath, "rb");
	if (!inf) {
		fclose(f);
		perror ("fopen read");
		return false;
	}
	ouf = fopen(outpath, "wb");
	if (!ouf) {
		perror ("fopen write");
		fclose(f);
		fclose(inf);
		return false;
	}
	hef = fopen(headerpath, "wb");
	if (!hef) {
		perror ("fopen write");
		fclose(f);
		fclose(inf);
		fclose(ouf);
		return false;
	}

	char word[100];
	int len;
	while ((len = getword (f, word, 99)) > 0) {
		if (!strcmp (word, "content-html")) {
			fprintf (ouf, "Content-type: text/html\n\n");
			continue;
		}
		
		if (word[0]=='0' && word[1]=='x') {
			unsigned long value;
			if (1 == sscanf (word+2, "%lx", &value)) {
				StackItem *im = StackItem_new();
				im->integer = value;
				im->type = STACKITEM_INTEGER;
				im->next = stack;
				stack = im;
			}
			continue;
		} else
		if (isdigit (word[0]) || 
		    ('-' == word[0] && isdigit(word[1]))) 
		{
			StackItem *im = StackItem_new();
			if (strchr (word, '.')) {
				double real;
				if (1 == sscanf (word, "%lf", &real)) {
					im->type = STACKITEM_REAL;
					im->real = real;
				} else
					break;
				im->next = stack;
				stack = im;
			} else {
				int integer;
				if (1 == sscanf (word, "%d", &integer)) {
					im->type = STACKITEM_INTEGER;
					im->integer = integer;
				} else
					break;
				im->next = stack;
				stack = im;
			}
			continue;
		}
		else
		if ('/' == word[0] && word[1]) {
			if (strchr (word+1, '\''))
				break;

			StackItem *im = StackItem_new();
			im->type = STACKITEM_SYMBOL;
			im->string = strdup(word+1);
			im->next = stack;
			stack = im;
			continue;
		}
		else
		if (!strcmp (word, "print") || !strcmp (word, "write") ||			    !strcmp (word, "writeln")) 
		{
			StackItem *tos = stack;
			if (!tos)
				break;
			stack = stack->next;

			if (tos->bold)
				fprintf (ouf, "<b>");
			if (tos->italic)
				fprintf (ouf, "<i>");
			if (tos->underline)
				fprintf (ouf, "<u>");
			if (tos->strikeout)
				fprintf (ouf, "<s>");
			if (tos->super)
				fprintf (ouf, "<sup>");
			if (tos->sub)
				fprintf (ouf, "<sub>");

			switch (tos->type) {
			case STACKITEM_INTEGER:
				fprintf (ouf, "%d", tos->integer);
				break;
			case STACKITEM_REAL:
				fprintf (ouf, "%lf", tos->real);
				break;
			case STACKITEM_STRING:
				fprintf (ouf, "%s", tos->string);
				break;
			case STACKITEM_SYMBOL:
				fprintf (ouf, "/%s", tos->string);
				break;
			}

			if (tos->sub)
				fprintf (ouf, "</sub>");
			if (tos->super)
				fprintf (ouf, "</sup>");
			if (tos->strikeout)
				fprintf (ouf, "</s>");
			if (tos->underline)
				fprintf (ouf, "</u>");
			if (tos->italic)
				fprintf (ouf, "</i>");
			if (tos->bold)
				fprintf (ouf, "</b>");

			StackItem_delete(tos);
			if (!strcmp (word, "writeln"))
				fprintf (ouf, "\n");
			continue;
		}
		else
		if (!strcmp (word, "++") || !strcmp (word, "--") ||
		    !strcmp (word, "!") || !strcmp (word, "neg")) {
			StackItem *tos = stack;
			if (!tos)
				break;
			if (tos->type == STACKITEM_INTEGER) {
intversion2:
				switch (word[0]) {
				case '+': tos->integer++; break;
				case '-': tos->integer--; break;
				case '!': tos->integer = !tos->integer; break;
				case 'n': tos->integer = -tos->integer; break;
				}
			} else
			if (tos->type == STACKITEM_REAL) {
				switch (word[0]) {
				case '+': tos->real += 1.0; break;
				case '-': tos->real -= 1.0; break;
				case '!': 
					tos->integer = tos->real;
					tos->type = STACKITEM_INTEGER;
					goto intversion2;
				}
			} 
			continue;
		}
		else
		if (!strcmp (word, "+") || !strcmp (word, "-") ||
		    !strcmp (word, "&") || !strcmp (word, "|") ||
		    !strcmp (word, "&&") || !strcmp (word, "||") ||
		    !strcmp (word, "==") || !strcmp (word, "!=") ||
		    !strcmp (word, "<") || !strcmp (word, ">") ||
		    !strcmp (word, "<=") || !strcmp (word, ">=") ||
		    !strcmp (word, "%") || !strcmp (word, "<<") ||
		    !strcmp (word, ">>") || !strcmp (word, "^") ||
		    !strcmp (word, "^^") || 
		    !strcmp (word, "*") || !strcmp (word, "/")) {
			StackItem *tos = stack;
			StackItem *nos;
			if (!tos)
				continue;
			nos = stack->next;
			if (!nos)
				break;
			stack = nos;
			if (tos->type != nos->type) {
				if (tos->type == STACKITEM_REAL &&
				    nos->type == STACKITEM_INTEGER) {
					nos->real = nos->integer;
				     	nos->type = STACKITEM_REAL;
				}
				else
				if (nos->type == STACKITEM_REAL &&
				    tos->type == STACKITEM_INTEGER) {
					tos->real = tos->integer;
				     	tos->type = STACKITEM_REAL;
				}
			}
			
			if (tos->type == STACKITEM_INTEGER) {
intversion:
				switch (word[0]) {
				case '=': nos->integer =
					nos->integer == tos->integer; 
					break;
				case '!': nos->integer =
					nos->integer != tos->integer; 
					break;
				case '+': nos->integer += tos->integer; break;
				case '-': nos->integer -= tos->integer; break;
				case '*': nos->integer *= tos->integer; break;
				case '&':
					if (word[1])
						nos->integer = 
							tos->integer &&
							nos->integer;
					else
						nos->integer &= tos->integer; 
					break;
				case '|': 
					if (word[1])
						nos->integer =
							tos->integer ||
							nos->integer;
					else
						nos->integer |= tos->integer; 
					break;
				case '^': 
					if (word[1])
						nos->integer =
							!!tos->integer ^
							!!nos->integer;
					else
						nos->integer ^= tos->integer; 
					break;
				case '<': 
					if (word[1] == '<')
						nos->integer <<= tos->integer; 
					else
					if (word[1] == '=')
						nos->integer =
						 nos->integer <= tos->integer; 
					else
						nos->integer =
						 nos->integer < tos->integer; 
					break;
				case '>': 
					if (word[1] == '>')
						nos->integer >>= tos->integer; 
					else
					if (word[1] == '=')
						nos->integer =
						 nos->integer >= tos->integer; 
					else
						nos->integer =
						 nos->integer > tos->integer; 
					break;
				case '/': 
					if (tos->integer) 
						nos->integer /= tos->integer; 
					else
						nos->integer = 0;
					break;
				case '%': 
					if (tos->integer) 
						nos->integer %= tos->integer; 
					else
						nos->integer = 0;
					break;
				}
			} else 
			if (tos->type == STACKITEM_REAL) {
				switch (word[0]) {
				case '=': nos->integer =
					nos->real == tos->real; 
					nos->type = STACKITEM_INTEGER;
					break;
				case '!': nos->integer =
					nos->real != tos->real; 
					nos->type = STACKITEM_INTEGER;
					break;
				case '+': nos->real += tos->real; break;
				case '-': nos->real -= tos->real; break;
				case '*': nos->real *= tos->real; break;
				case '<':
					if (word[1] == '=') {
						nos->integer =
						  nos->real <= tos->real; 
						nos->type = STACKITEM_INTEGER;
printf ("nosreal %lf tosreal %lf nosint %d\n",
nos->real,tos->real, nos->integer);
					} else
					if (!word[1]) {
						nos->integer =
						  nos->real < tos->real; 
						nos->type = STACKITEM_INTEGER;
					} else
						goto intversion;
					break;
				case '>':
					if (word[1] == '=') {
						nos->integer =
						  nos->real >= tos->real; 
						nos->type = STACKITEM_INTEGER;
					} else
					if (!word[1]) {
						nos->integer =
						  nos->real > tos->real; 
						nos->type = STACKITEM_INTEGER;
					} else
						goto intversion;
					break;
				case '/': 
					if (tos->real) 
						nos->real /= tos->real; 
					else
						nos->real = 0;
					break;
				case '&':
				case '|': 
				case '^': 
				case '%': 
					/* Convert to integer */
					tos->integer = tos->real;
					nos->integer = nos->real;
					tos->type = nos->type =
						STACKITEM_INTEGER;
					goto intversion;
				}
				continue;
			} else 
			if (tos->type == STACKITEM_STRING) {
				int noslen = strlen(nos->string);
				char *s = malloc (noslen + 2 + 
					strlen(tos->string));
				strcpy (s, nos->string);
				strcpy (s + noslen, tos->string);
				free (nos->string);
				nos->string = s;
			} else 
				break;

			StackItem_delete(tos);
			continue;
		}
		else
		if (!strcmp (word,"version")) {
			StackItem *im = StackItem_new();
			im->string = strdup (VERSION);
			im->type = STACKITEM_STRING;
			im->next = stack;
			stack = im;
		}
		else
		if (!strcmp (word,"exch")) {
			StackItem *nos;
			if (!stack)
				continue;
			nos = stack->next;
			if (!nos)
				continue;
			stack->next = stack;
			nos->next = stack;
			stack = nos;
		}
		else
		if (!strcmp (word,"style")) {
			if (stack && stack->type == STACKITEM_STRING)
				stack->type = STACKITEM_STYLE;
			// XX might want to validate it
		}
		else
		if (!strcmp (word,"nostyle")) {
			StackItem *im = StackItem_new ();
			im->next = stack;
			stack = im;
			im->type = STACKITEM_STYLE;
			im->string = NULL;
		}
		else
		if (!strcmp (word,"div-begin")) {
			StackItem *style = stack;
			if (!style || style->type != STACKITEM_STYLE)
				continue;
			stack = stack->next;
			fprintf (ouf, "<div style=\"%s\">", style->string);
			StackItem_delete (style);
		}
		else
		if (!strcmp (word,"div-end"))
			fprintf (ouf, "</div>\n");
		else
		if (!strcmp (word,"table-begin")) {
			fprintf (ouf, "<table border=1>\n");
		}
		else
		if (!strcmp (word,"table-end")) {
			fprintf (ouf, "</table>\n");
		}
		else
		if (!strcmp (word,"table-row")) {
			fprintf (ouf, "<tr>");
		}
		else
		if (!strcmp (word,"table-column")) {
			fprintf (ouf, "<td>\n");
		}
		else
		if (!strcmp (word,"table-hdr")) {
			fprintf (ouf, "<th>\n");
		}
		else
		if (!strcmp (word,"list-begin")) {
			fprintf (ouf, "<ul>\n");
		}
		else
		if (!strcmp (word,"list-item")) {
			fprintf (ouf, "<li>\n");
		}
		else
		if (!strcmp (word,"list-end")) {
			fprintf (ouf, "</ul>\n");
		}
		else
		if (!strcmp (word,"numbered-begin")) {
			fprintf (ouf, "<ol>\n");
		}
		else
		if (!strcmp (word,"numbered-item")) {
			fprintf (ouf, "<li>\n");
		}
		else
		if (!strcmp (word,"numbered-end")) {
			fprintf (ouf, "</ol>\n");
		}
		else
		if (!strcmp (word,"description-begin")) {
			fprintf (ouf, "<dl>\n");
		}
		else
		if (!strcmp (word,"description-title")) {
			fprintf (ouf, "<dt>\n");
		}
		else
		if (!strcmp (word,"description-contents")) {
			fprintf (ouf, "<dd>\n");
		}
		else
		if (!strcmp (word,"description-end")) {
			fprintf (ouf, "</ol>\n");
		}
		else
		if (!strcmp (word,"directory-begin")) {
			fprintf (ouf, "<dir>\n");
		}
		else
		if (!strcmp (word,"directory-end")) {
			fprintf (ouf, "</dir>\n");
		}
		else
		if (!strcmp (word,"bold-begin")) {
			fprintf (ouf, "<b>\n");
		}
		else
		if (!strcmp (word,"bold-end")) {
			fprintf (ouf, "</b>\n");
		}
		else
		if (!strcmp (word,"italic-begin")) {
			fprintf (ouf, "<i>\n");
		}
		else
		if (!strcmp (word,"italic-end")) {
			fprintf (ouf, "</i>\n");
		}
		else
		if (!strcmp (word,"strike-begin")) {
			fprintf (ouf, "<s>\n");
		}
		else
		if (!strcmp (word,"strike-end")) {
			fprintf (ouf, "</s>\n");
		}
		else
		if (!strcmp (word,"underline-begin")) {
			fprintf (ouf, "<u>\n");
		}
		else
		if (!strcmp (word,"underline-end")) {
			fprintf (ouf, "</u>\n");
		}
		else
		if (!strcmp (word,"link")) {
			StackItem *tos, *nos;
			tos = stack;
			if (!tos)
				continue;
			nos = tos->next;
			if (!nos)
				continue;
			if (tos->type != STACKITEM_STRING ||
			    nos->type != STACKITEM_STRING)
				continue;

			stack = nos->next;
			fprintf (ouf, "<a href=\"%s\">%s</a>\n", 
				nos->string,
				tos->string);
			StackItem_delete (tos);
			StackItem_delete (nos);
		}
		else
		if (!strcmp (word,"set-cookie")) {
			StackItem *tos, *nos;
			tos = stack;
			if (!tos)
				continue;
			nos = tos->next;
			if (!nos)
				continue;
			if (tos->type != STACKITEM_STRING ||
			    nos->type != STACKITEM_STRING)
				continue;

			stack = nos->next;
			fprintf (hef, "Set-cookie: %s=%s\n",
				nos->string,
				tos->string);
			StackItem_delete (tos);
			StackItem_delete (nos);
		}
		else
		if (!strcmp (word,"bold")) {
			if (stack)
				stack->bold = true;
		}
		else
		if (!strcmp (word,"italic")) {
			if (stack)
				stack->italic= true;
		}
		else
		if (!strcmp (word,"underline")) {
			if (stack)
				stack->underline= true;
		}
		else
		if (!strcmp (word,"strikeout")) {
			if (stack)
				stack->strikeout= true;
		}
		else
		if (!strcmp (word,"superscript")) {
			if (stack)
				stack->super= true;
		}
		else
		if (!strcmp (word,"subscript")) {
			if (stack)
				stack->sub= true;
		}
		else
		if (!strcmp (word,"break")) {
			fprintf (ouf, "<br>\n");
		}
		else
		if (!strcmp (word,"hline")) {
			fprintf (ouf, "<hr>\n");
		}
		else
		if (!strcmp (word,"par")) {
			fprintf (ouf, "<p>\n");
		}
		else
		if (!strcmp (word,"def")) {
		}
		else {
			// Assume it's just a string
			StackItem *im = StackItem_new();
			im->string = strdup (word);
			im->type = STACKITEM_STRING;
			im->next = stack;
			stack = im;
		}
	}

	fclose(f);
	fclose(inf);
	fclose(ouf);
	fclose(hef);
	return 0;
}
#endif


bool process_script (int fd, char *localpath, char type)
{
	char cmd [4096];
	char inputpath [30];
	char outputpath [30];

	maketempfile (inputpath);
	maketempfile (outputpath);

	FILE *f = fopen (inputpath, "w");
	fprintf (f, "foo=1\n");
	fclose (f);

	char *ws=NULL;
	switch (type) {
	case 1: ws = "perl"; break;
	case 2: ws = "php"; break;
	case 3: ws = "python"; break;
	default:
		return false;
	}
	if (type==2)
		sprintf (cmd, "%s -f %s %s > %s", ws, localpath, inputpath, outputpath);
	else
		sprintf (cmd, "%s %s < %s > %s", ws, localpath, inputpath, outputpath);
	system(cmd);

	unlink (inputpath);

	if (!send_results (fd, outputpath, NULL))  {
		unlink (outputpath);
		log_write ("\tERROR: script results send failed\n");
		http_error (fd, HTTP_ERR_INTERNAL);
	}
	
	unlink (outputpath);
	log_store ();
	return myexit(0);
}


#ifdef SUPPORT_SY
// Process any Sy program
//
bool process_sy (int fd, char *localpath)
{
	char inputpath [30];
	char outputpath [30];
	char headerpath [30];

	maketempfile (inputpath);
	maketempfile (outputpath);
	maketempfile (headerpath);

	FILE *f = fopen (inputpath, "w");
	fprintf (f, "foo=1\n");
	fclose (f);

	process_sy_core (fd, localpath, inputpath, outputpath, headerpath);
printf ("finished processing Sy code\n");

	unlink (inputpath);

	if (!send_results (fd, outputpath, headerpath)) {
		unlink (outputpath);
		unlink (headerpath);
		log_write ("\tERROR: Sy send failed\n");
		http_error (fd, HTTP_ERR_INTERNAL);
	}
	
	unlink (outputpath);
	unlink (headerpath);
	log_store ();
	return myexit(0);
}
#endif
